@@ -32,6 +32,7 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <ctype.h>
+#include <glib.h>
 
 /* ============================ _fmt_ptrn_t_valid () ======================= */
 static gboolean _fmt_ptrn_t_valid(const fmt_ptrn_t * x)
@@ -117,12 +118,6 @@ void fmt_ptrn_enqueue_parse_errmsg(fmt_ptrn_t * x, const char *msg, ...)
 	assert(_fmt_ptrn_t_valid(x));
 }
 
-/* ============================ mstack_t =================================== */
-typedef struct mstack_t {
-	modifier_t data[STACK_MAX_ITEMS];
-	int size;
-} mstack_t;
-
 /* ============================ _modifier_t_valid () ======================= */
 static gboolean _modifier_t_valid(const modifier_t * m)
 {
@@ -130,74 +125,6 @@ static gboolean _modifier_t_valid(const modifier_t * m)
 	return TRUE;
 }
 
-/* ============================ _mstack_t_valid () ========================= */
-static gboolean _mstack_t_valid(const mstack_t * s)
-{
-	if (s == NULL && s->size != 0)
-		return FALSE;
-	return TRUE;
-}
-
-/* ============================ _stack_init () ============================= */
-static void _stack_init(mstack_t * s)
-{
-	s->size = 0;
-	assert(_mstack_t_valid(s));
-}
-
-/* ============================ _stack_push () ============================= */
-static gboolean _stack_push(fmt_ptrn_t * x, mstack_t * s,
-			    const modifier_t data)
-{
-	gboolean fnval = FALSE;
-
-	assert(_fmt_ptrn_t_valid(x));
-	assert(_mstack_t_valid(s));
-	assert(_modifier_t_valid(&data));
-	if (s->size == STACK_MAX_ITEMS) {
-		fmt_ptrn_enqueue_parse_errmsg(x, "%s: %ld: more than %d modifiers",
-				     x->template_path, x->line_num,
-				     STACK_MAX_ITEMS);
-		fnval = FALSE;
-	} else {
-		s->data[s->size++] = data;
-		fnval = TRUE;
-	}
-
-	assert(_fmt_ptrn_t_valid(x));
-	assert(_mstack_t_valid(s));
-	return fnval;
-}
-
-/* ============================ _stack_pop () ============================== */
-static gboolean _stack_pop(mstack_t * s, modifier_t * data)
-{
-	gboolean fnval = FALSE;
-	assert(_mstack_t_valid(s));
-	assert(_modifier_t_valid(data));
-	if (!s->size)
-		fnval = FALSE;
-	else {
-		*data = s->data[--s->size];
-		fnval = TRUE;
-	}
-
-	assert(_mstack_t_valid(s));
-	assert(_modifier_t_valid(data));
-	return fnval;
-}
-
-/* ============================ _stack_contains () ========================= */
-static int _stack_contains(const mstack_t s, const char *n)
-{
-	int i;
-	assert(_mstack_t_valid(&s));
-	for (i = 0; i < s.size; i++)
-		if (s.data[i].id == n)
-			return 1;
-	return 0;
-}
-
 /* ============================ fmt_ptrn_update_kv_p () ==================== */
 void fmt_ptrn_update_kv_p(fmt_ptrn_t * x, const pair_t * p)
 {
@@ -386,11 +313,11 @@ modifier_fns_t *fmt_ptrn_get_modifier(fmt_ptrn_t *x, const char *id)
 
 /* ============================ _read_modifier () ========================== */
 static gboolean _read_modifier(fmt_ptrn_t * x, char **ptrn,
-			       mstack_t * modifier)
+			       GQueue *modifier)
 {
 	int i = 0;
 	size_t len;
-	modifier_t m;
+	modifier_t *m;
 	modifier_fns_t *mod_fn;
 	gboolean fnval = FALSE;
 	const char *id;
@@ -398,8 +325,8 @@ static gboolean _read_modifier(fmt_ptrn_t * x, char **ptrn,
 	assert(_fmt_ptrn_t_valid(x));
 	assert(ptrn != NULL);
 	assert(*ptrn != NULL);
-	assert(_mstack_t_valid(modifier));
 
+	m = g_new (modifier_t, 1);
 
 	/* FIXME: slow and awkward? */
 	len = strlen(*ptrn);
@@ -410,32 +337,30 @@ static gboolean _read_modifier(fmt_ptrn_t * x, char **ptrn,
 	mod_fn = fmt_ptrn_get_modifier(x, id);
 	if (NULL != mod_fn) {
 		*ptrn += strlen(id) + (mod_fn->has_arg ? 0 : 1);
-		m.id = id;
-		m.fn = mod_fn;
+		m->id = id;
+		m->fn = mod_fn;
 		if (mod_fn->has_arg)
-			_read_modifier_arg(x, ptrn, &m);
+			_read_modifier_arg(x, ptrn, m);
 		else
-			*m.arg = 0x00;
-		_stack_push(x, modifier, m);
+			*(m->arg) = 0x00;
+		g_queue_push_head (modifier, m);
 		fnval = TRUE;
 	}
 
 	assert(_fmt_ptrn_t_valid(x));
 	assert(ptrn != NULL);
 	assert(*ptrn != NULL);
-	assert(_mstack_t_valid(modifier));
 
 	return fnval;
 }
 
 /* ============================ _read_modifiers () ========================= */
 static void _read_modifiers(fmt_ptrn_t * x,
-			    char **ptrn, mstack_t * modifier)
+			    char **ptrn, GQueue *modifier)
 {
 	assert(_fmt_ptrn_t_valid(x));
 	assert(ptrn != NULL);
 	assert(*ptrn != NULL);
-	assert(_mstack_t_valid(modifier));
 
 	while (_read_modifier(x, ptrn, modifier)) {
 		/* NOOP. */
@@ -444,7 +369,6 @@ static void _read_modifiers(fmt_ptrn_t * x,
 	assert(_fmt_ptrn_t_valid(x));
 	assert(ptrn != NULL);
 	assert(*ptrn != NULL);
-	assert(_mstack_t_valid(modifier));
 
 	return;
 }
@@ -480,30 +404,29 @@ static void _read_key(fmt_ptrn_t * x, char *key, char **p)
 
 /* ============================ _apply_modifiers () ======================== */
 static void _apply_modifiers(fmt_ptrn_t * x,
-			     buffer_t * str, mstack_t * modifier)
+			     buffer_t * str, GQueue *modifier)
 {
-	modifier_t m;
+	modifier_t *m;
 
 	assert(_fmt_ptrn_t_valid(x));
 	assert(buffer_t_valid(str));
-	assert(_mstack_t_valid(modifier));
 
 	if (buffer_len(str) > 0)	/* error should have been queued elsewhere */
-		while (_stack_pop(modifier, &m)) {
-			if (m.fn->fn)
-				if (!m.fn->fn(str, x, m.arg))
+		while ((m = g_queue_pop_head (modifier))) {
+			if (m->fn->fn)
+				if (!m->fn->fn(str, x, m->arg))
 					fmt_ptrn_enqueue_parse_errmsg(x,
 							     "%s: %ld: error applying %s modifier to %s",
 							     x->
 							     template_path,
 							     x->line_num,
-							     m.id,
+							     m->id,
 							     str->data);
-							     }
+			g_free (m);
+		}
 
 	assert(_fmt_ptrn_t_valid(x));
 	assert(buffer_t_valid(str));
-	assert(_mstack_t_valid(modifier));
 }
 
 /* ============================ _lookup () ================================= */
@@ -577,26 +500,28 @@ static void _read_literal(fmt_ptrn_t *x, char *str, buffer_t * buf)
 static void _handle_fmt_str(fmt_ptrn_t * x, char **p)
 {
 	/* format string -> %(<modifier_0> ... <modifier_n> <key>:<alt>) */
-	mstack_t modifier;
+	GQueue *modifier;
 	char key[KEY_LEN + 1];
 
 	assert(_fmt_ptrn_t_valid(x));
 	assert(p != NULL);
 	assert(*p != NULL);
 
-	_stack_init(&modifier);
+	modifier = g_queue_new ();
 	*p += 2;		/* Skip "%(". */
-	_read_modifiers(x, p, &modifier);
-	if (_stack_contains(modifier, "#")) {
+	_read_modifiers(x, p, modifier);
+	/* FIXME: What is this?
+	if (_stack_contains(modifier, "#")) { */
+	if (0) {
 		/* NOOP. */
 	} else {
 		_read_key(x, key, p);
 		if (_is_literal(x, key)) {
 			_read_literal(x, key, &x->lookup_buf);
-			_apply_modifiers(x, &x->lookup_buf, &modifier);
+			_apply_modifiers(x, &x->lookup_buf, modifier);
 		} else if (_lookup(x, key, &x->lookup_buf)) {
 			_eat_alternate(x, p);
-			_apply_modifiers(x, &x->lookup_buf, &modifier);
+			_apply_modifiers(x, &x->lookup_buf, modifier);
 		} else
 			_read_alternate(x, p, &x->lookup_buf);
 		if (buffer_len(&x->lookup_buf))
@@ -606,6 +531,8 @@ static void _handle_fmt_str(fmt_ptrn_t * x, char **p)
 	if (**p)
 		(*p)++;		/* Skip ')'. */
 
+	g_queue_free (modifier);
+
 	assert(_fmt_ptrn_t_valid(x));
 	assert(p != NULL);
 	assert(*p != NULL);
@@ -654,15 +581,18 @@ GList *fmt_ptrn_get_keys(fmt_ptrn_t * x, const char *p)
 	while (*pattern != 0x00) {
 		if (*pattern == '%' && *(pattern + 1) == '(') {
 			/* format string -> %(<modifier_0> ... <modifier_n> <key>:<alt>) */
-			mstack_t modifier;
+			GQueue *modifier;
 			char key[KEY_LEN + 1];
 
 			assert(*pattern != 0x00);
 
-			_stack_init(&modifier);
+			modifier = g_queue_new ();
+
 			pattern += 2;		/* Skip "%(". */
-			_read_modifiers(x, &pattern, &modifier);
-			if (_stack_contains(modifier, "#")) {
+			_read_modifiers(x, &pattern, modifier);
+			/* FIXME: see above duplicate:
+			if (_stack_contains(modifier, "#")) { */
+			if (0) {
 				/* NOOP. */
 			} else {
 				_read_key(x, key, &pattern);
@@ -676,6 +606,8 @@ GList *fmt_ptrn_get_keys(fmt_ptrn_t * x, const char *p)
 			if (*pattern)
 				pattern++;		/* Skip ')'. */
 
+			g_queue_free (modifier);
+
 		}
 		pattern++;
 	}
